import Dist
reload(Dist)
from Dist import *

# State machines, when they're initialized, they need to bind the following
# variables:
#    self.name: string
#    self.initialState: initialState
#    self.inputs: list(input)
#    self.transitionFunction: state x input -> state
#    self.observationFunction: state -> output
#    self.terminationTest: state -> Boolean

class SM:
    def step(self, input=None):
        print self.name, " oldState: ", self.currentState, " input: ", input
        self.currentState = self.transitionFunction(self.currentState, input)
        output = self.outputFunction(self.currentState)
        print self.name, " newState: ", self.currentState, " output: ", output
        return output

    def done(self):
        return self.terminationTest(self.currentState)
    
    def state(self):
        return self.currentState

    def reset(self):
        self.currentState = self.initialState

    def run(self):
        self.reset()
        while not self.done():
            self.step()

def runSM(sim):
   while not sim.done():
       sim.step()

class PrimitiveSM(SM):
    def __init__(self, name, transitionfn, outputfn, initState,
                 terminationfn = None, inputs = None):
        self.name = name
        self.transitionFunction = transitionfn
        self.outputFunction = outputfn
        if terminationfn == None:
            self.terminationTest = lambda x: False
        else:
            self.terminationFunction = terminationfn
        self.initialState = initState
        self.currentState = initState
        self.inputs = inputs

class SerialSM(SM):
    def __init__(self, sm1, sm2):
        self.m1 = sm1
        self.m2 = sm2
        self.inputs = sm1.inputs
        def transitionFunction((s1, s2), a1):
            if s == None:
                return None
            news1 = sm1.transitionFunction(s1, a1)
            o1 = sm1.outputFunction(news1)
            news2 = sm2.transitionFunction(s2, o1)
            return (news1, news2)
        self.transitionFunction = transitionFunction
        def outputFunction((s1, s2)):
            return sm2.outputFunction(s2)
        self.outputFunction = outputFunction
        self.initialState = (sm1.initialState, sm2.initialState)

    def step(self, input):
        if not self.done():
            out1 = self.m1.step(input)
            if self.m1.done():     
                return 'Done'
            return self.m2.step(out1)

    def state(self):
        return [self.m1.state(), self.m2.state()]

    def done(self):
        return self.m1.done() or self.m2.done()

    def reset(self):
        self.m1.reset()
        self.m2.reset()

class ParallelSM (SM):
    def __init__(self, sm1, sm2):
        self.m1 = sm1
        self.m2 = sm2
        self.inputs = [(a1, a2) for a1 in sm1.inputs for a2 in sm2.inputs]
        def transitionFunction(s, (a1, a2)):
            if s == None:
                return None
            (s1, s2) = s
            news1 = self.m1.transitionFunction(s1, a1)
            news2 = self.m2.transitionFunction(s2, a2)
            if news1 and news2: 
                return (news1, news2)
            else:
                return None
        self.transitionFunction = transitionFunction
        
    def step(self, (i1, i2)):
        if not self.done():
            return [self.m1.step(i1), self.m2.step(i2)]

    def state(self):
        return [self.m1.state(), self.m2.state()]
    
    def done(self):
        return self.m1.done() or self.m2.done()

    def reset(self):
        self.m1.reset()
        self.m2.reset()

class FeedbackSM (SM):
    def __init__(self, tsm, initInput):
        self.m = tsm
        self.initInput = initInput
        self.feedback = initInput
    
    def step(self, input=None):
        if not self.done():
            self.feedback = self.m.step(self.feedback)
        
    def state(self):
        return [feedback, self.m.state()]

    def done(self):
        return self.m.done()

    def reset(self):
        self.m.reset()
        self.feedback = self.initInput

class PrimitiveStochasticSM (PrimitiveSM):
    def step(self, input=None):
        print self.name, " oldState: ", self.currentState, " input: ", input
        self.currentState = \
                       self.transitionFunction(self.currentState,input).draw()
        output = self.outputFunction(self.currentState).draw()
        print self.name, " newState: ", self.currentState, " output: ", output
        return output

def simulatorSM(m1, m2, initInput):
    return FeedbackSM(SerialSM(m1, m2), initInput)




