import DrawingWindowStandalone
reload(DrawingWindowStandalone)
from DrawingWindowStandalone import *

import probcolors
reload(probcolors)
from probcolors import *

import SM
reload(SM)
from SM import *

import ButtonWindow
reload(ButtonWindow)
from ButtonWindow import *

# world is a SM.  It has to support the following interface:
# - state is a list of two positive integers, bounded size
# - action is a list of two integers
# - fsm.xdim and fsm.ydim give the max ranges of state variables
# - fsm.state gives the current state
# - fsm.transitionFunction and fsm.outputFunction return distributions
# - observations are positive integers

class GridStateEstimator:
    def __init__(self, world, window = None):
        self.verbose = True
        self.world = world
        self.belief = arrayDiv(ones2Float(world.xdim, world.ydim),
                               (world.xdim * world.ydim))
        print "Initial belief before any actions"
        self.printBelief()
        if window:
            self.window = window
        else:
            self.window = DrawingWindow(world.xdim*50, world.ydim*50+10,
                                        -0.2, world.xdim+0.2, -0.2,
                                        world.ydim + 0.2, "Belief")
    def reset(self):
        self.belief = arrayDiv(ones2Float(self.world.xdim, self.world.ydim),
                               (self.world.xdim * self.world.ydim))

    def printBelief(self):
        for y in range(self.world.ydim-1, -1, -1):
            string = ""
            for x in range(self.world.xdim):
                string +=  "%8.4f"%self.belief[x][y]
            print string

    def actionUpdate(self, (ax, ay)):
        newbelief = zeros2Float(self.world.xdim, self.world.ydim)
        # For each old belief state, move the probability mass
        # according to the action.  
        for x in range(self.world.xdim):
            for y in range(self.world.ydim):
                result = self.world.transitionFunction((x,y),(ax,ay))
                for ((nx,ny),np) in result.sampleSpace:
                    newbelief[nx][ny] += self.belief[x][y]*np
        self.belief = newbelief
        if self.verbose:
            print "Updated belief after action", (ax, ay)
            self.printBelief()

    def observationUpdate(self, obs):
        # For each old belief state, multiply in the probability of
        # making this observation in that state.  Then renormalize.
        total = 0.0
        for x in range(self.world.xdim):
            for y in range(self.world.ydim):
                sensorDist = self.world.outputFunction([x,y])
                self.belief[x][y] *= sensorDist.prob(obs)
                total += self.belief[x][y]
        self.belief = arrayDiv(self.belief, total)
        if self.verbose:
            print "Updated belief after observation", obs
            self.printBelief()

    def drawBelief(self):
        unifP = 1.0/(self.world.xdim * self.world.ydim)
        for x in range(self.world.xdim):
            for y in range(self.world.ydim):
                self.window.drawRect((x + 0.2, y + 0.2), (x + 0.8, y + 0.8),
                                     color = beliefColor(self.belief[x][y],
                                                         unifP))
        
class ColorRobotWorld(PrimitiveStochasticSM):
    # colors is a list of lists of xy pairs.  First list is locations of
    # black squares, second is list of red squares, then green, then blue
    # squares are white by default
    
    # robot is initial location of robot

    # motion model takes state (x,y), action (dx, dy), and bounds (mx, my) and
    # returns a sparse distribution over next states

    # sensor model takes nominal output and returns a sparse distribution over
    # outputs

    # remember the transition and output distribution functions, because we may
    # want to do something analytic with them later on (not just draw from the
    # distributions for simulation)

    def __init__(self, xdim, ydim, colors, robot, sensorModel, motionModel):
        self.xdim = xdim
        self.ydim = ydim
        self.terminationTest = lambda s: False
        self.grid = makeGrid(xdim, ydim, colors)
        self.transitionDistFn = lambda s, a: motionModel(s, a, (xdim, ydim))
        self.outputDistFn = lambda (x,y): sensorModel(self.grid[x][y])

        PrimitiveStochasticSM.__init__(self, "ColorWorld",
                                        self.transitionDistFn,
                                        self.outputDistFn, robot)

class ColorRobotWorldWithDisplay(ColorRobotWorld):
    def __init__(self, xdim, ydim, colors, robot, sensorModel,
                 motionModel, window = None):
        ColorRobotWorld.__init__(self, xdim, ydim, colors, robot,
                                 sensorModel, motionModel)
        if window:
            self.window = window
        else:
            self.window = DrawingWindow(xdim*50, ydim*50+10, -0.2, xdim+0.2, \
                                        -0.2, ydim + 0.2, "World")
        self.draw()
        
    def step(self, input):
        obs = ColorRobotWorld.step(self, input)
        self.draw()
        return obs

    def reset(self):
        PrimitiveStochasticSM.reset(self)
        self.draw()

    def drawWorld(self):
        for x in range(self.xdim):
            for y in range(self.ydim):
                self.window.drawRect((x, y), (x+1, y+1),
                                     color = colors[self.grid[x][y]])

    def drawRobot(self):
        (rx,ry) = self.state()
        self.window.drawRect((rx + 0.4, ry+0.4), (rx + 0.6, ry + 0.6),
                             color = 'orange')

    def draw(self):
        self.drawWorld()
        self.drawRobot()


# This is an FSM that keeps a belief state updated over time based on
# actions and observations.  It shows the belief state to a user and
# asks the user to pick an action, and returns it.

class GridStateEstimatorUI(PrimitiveSM):
    def __init__(self, fsm, actionNames, actionVals, SEwindow=None,
                 buttonFrame = None, top=None):
        self.lastAction = (0,0)
        self.initialState = fsm.initialState
        self.se = GridStateEstimator(fsm, SEwindow)
        self.se.drawBelief()
        self.buttonDiag = ButtonDialog(top, buttonFrame,
                                       actionNames, actionVals, True)
        self.name = "GridStateEstimatorUI"

    def step(self, observation):
        self.se.actionUpdate(self.lastAction)
        self.se.observationUpdate(observation)
        self.se.drawBelief()
        self.lastAction = self.buttonDiag.step(colors[observation])
        return self.lastAction

    def done(self):
        return self.buttonDiag.done()

    def reset(self):
        self.lastAction = (0,0)
        self.se.reset()
        self.se.drawBelief()
        self.buttonDiag.reset()

# This is an FSM that keeps a belief state updated over time based on
# actions and observations.  It shows the belief state to a user and
# asks the user to pick an action, and returns it.
import time
import sys
class GridStateEstimatorTextUI(PrimitiveSM):
    def __init__(self, fsm, actionNames, actionVals, top, SEwindow=None):
        self.lastAction = (0,0)
        self.initialState = fsm.initialState
        self.top = top
        self.se = GridStateEstimator(fsm, SEwindow)
        self.se.drawBelief()
        self.name = "GridStateEstimatorUI"
        self.actionNames = actionNames
        self.actionDict = makeDict(actionNames, actionVals)
        self.actionDict['done'] = 'done'

    def step(self, observation):
        self.se.actionUpdate(self.lastAction)
        self.se.observationUpdate(observation)
        self.se.drawBelief()
        self.top.update()
        self.lastAction = self.promptForAction(colors[observation])
        return self.lastAction

    def promptForAction(self, obs):
        print "Observed: ", obs
        print "Action (one of " + str(self.actionNames) + " or 'done') : "
        a = sys.stdin.readline()[:-1]
        while not self.actionDict.has_key(a):
            print "Action (one of " + str(self.actionNames) + " or 'done') : "
            a = sys.stdin.readline()[:-1]
        return self.actionDict[a]

    def done(self):
        return self.lastAction == "done"

    def reset(self):
        self.lastAction = (0,0)
        self.se.reset()
        self.se.drawBelief()

######################################################################
##
##   One big UI pane  (pain?)
##
######################################################################

import DrawingFrame
reload(DrawingFrame)
from DrawingFrame import *

def makeGridSimWithButtons(xdim, ydim, colors, robotLoc, sensorModel, motionModel):
    top = Tk() 
    top.wm_title("Grid World:  Actual state and Belief State")
    beliefWorldFrame = Frame(top)
    beliefWorldFrame.pack(side=TOP)
    beliefWindow = DrawingFrame(xdim*50,ydim*50,-0.2,xdim+0.2,-0.2,ydim+0.2,
                                beliefWorldFrame)
    worldWindow = DrawingFrame(xdim*50,ydim*50,-0.2,xdim+0.2,-0.2,ydim+0.2,
                               beliefWorldFrame)
    worldWindow.canvas.pack(side=LEFT)
    beliefWindow.canvas.pack(side=LEFT)
    world = ColorRobotWorldWithDisplay(xdim, ydim, colors, robotLoc,
                                       sensorModel, motionModel, worldWindow)
    actionStrings = ['North', 'South', 'East', 'West', 'Stay']
    actions = [[0,1],[0,-1],[1,0],[-1,0],[0,0]]
    gui =  GridStateEstimatorUI(world, actionStrings, actions,
                                beliefWindow, top, top)
    gui.buttonDiag.frame.pack(side=BOTTOM)
    return simulatorSM(world, gui, (0,0))

def makeGridSim(xdim, ydim, colors, robotLoc, sensorModel,
                         motionModel):
    top = Tk() 
    top.wm_title("Grid World:  Actual state and Belief State")
    beliefWorldFrame = Frame(top)
    beliefWorldFrame.pack(side=TOP)
    beliefWindow = DrawingFrame(xdim*50,ydim*50,-0.2,xdim+0.2,-0.2,ydim+0.2,
                                beliefWorldFrame)
    worldWindow = DrawingFrame(xdim*50,ydim*50,-0.2,xdim+0.2,-0.2,ydim+0.2,
                               beliefWorldFrame)
    worldWindow.canvas.pack(side=LEFT)
    beliefWindow.canvas.pack(side=LEFT)
    world = ColorRobotWorldWithDisplay(xdim, ydim, colors, robotLoc,
                                       sensorModel, motionModel, worldWindow)
    actionStrings = ['n', 's', 'e', 'w', 'stay']
    actions = [[0,1],[0,-1],[1,0],[-1,0],[0,0]]
    gui =  GridStateEstimatorTextUI(world, actionStrings, actions,
                                    top, beliefWindow)
    return simulatorSM(world, gui, (0,0))


######################################################################
##
## utils
##
######################################################################
        
        
colors = {0: 'white', 1: 'black', 2: 'red', 3:'green', 4:'blue'}
inverseColors = {'white':0, 'black':1, 'red':2, 'green':3, 'blue':4}

def beliefColor(p, uniformP):
    return probToPyColor(p, uniformP)

def makeGrid(xdim, ydim, (black, red, green, blue)):
    grid = zeros2Int(xdim, ydim)
    for (x,y) in black:
        grid[x][y] = inverseColors['black']
    for (x,y) in red:
        grid[x][y] = inverseColors['red']
    for (x,y) in green:
        grid[x][y] = inverseColors['green']
    for (x,y) in blue:
        grid[x][y] = inverseColors['blue']
    return grid

def clip(v, vmin, vmax):
    return max(vmin, min(vmax, v))

##########################################################################
###
###     So we don't need Numeric
###     
##########################################################################
def makeArray(xdim, ydim, init):
    result = []
    for x in range(xdim):
        result = result + [[init]*ydim]
    return result

def ones2Float(xdim, ydim):
    return makeArray(xdim, ydim, 1.0)

def zeros2Float(xdim, ydim):
    return makeArray(xdim, ydim, 0.0)

def zeros2Int(xdim, ydim):
    return makeArray(xdim, ydim, 0)

def arrayDiv(arr, x):
    return map(lambda row: map(lambda elt: elt / x, row), arr)


######################################################################
##
##  testing

def perfectSensorModel(trueColor):
    return Dist([[trueColor, 1.0]])

def perfectMotionModel((rx, ry), (ax, ay), (xmax, ymax)):
    def cx(a): return clip(a, 0, xmax-1)
    def cy(a): return clip(a, 0, ymax-1)
    return Dist([[[cx(rx+ax),cy(ry+ay)], 1.0]])

## Note that this is written for a two-dimensional grid.  The behavior on a one-dimensional 
## grid may be a little unintuitive because of clipping.
def noisyMotionModel((rx, ry), (ax, ay), (xmax, ymax)):
    def cx(a): return clip(a, 0, xmax-1)
    def cy(a): return clip(a, 0, ymax-1)
    noiseProb = 0.2
    p = noiseProb/8.0
    return Dist((((cx(rx+ax),cy(ry+ay)), 1 - noiseProb),
                 ((cx(rx+ax+1),cy(ry+ay+1)), p), 
                 ((cx(rx+ax+1),cy(ry+ay)), p),
                 ((cx(rx+ax+1),cy(ry+ay-1)), p),
                 ((cx(rx+ax),cy(ry+ay+1)), p),
                 ((cx(rx+ax),cy(ry+ay-1)), p),
                 ((cx(rx+ax-1),cy(ry+ay+1)), p),
                 ((cx(rx+ax-1),cy(ry+ay)), p),
                 ((cx(rx+ax-1),cy(ry+ay-1)), p)))

## This only models uncertainty for motion in X
def noisyMotionModelX((rx, ry), (ax, ay), (xmax, ymax)):
    def cx(a): return clip(a, 0, xmax-1)
    def cy(a): return clip(a, 0, ymax-1)
    noiseProb = 0.2
    p = noiseProb/2.0
    return Dist((((cx(rx+ax),cy(ry+ay)), 1 - noiseProb),
                 ((cx(rx+ax+1),cy(ry+ay)), p),
                 ((cx(rx+ax-1),cy(ry+ay)), p)))

def noisySensorModel(trueColor):
    ## the default value is 0.2/4 = 0.05
    values = [[0, 0.05], [1, 0.05], [2, 0.05], [3, 0.05], [4, 0.05]]
    ## the designated color gets 0.8
    values[trueColor] = [trueColor, 0.8]
    return Dist(values)

## An example of a Dist
## Dist( [ [ 1, 0.1 ], [2, 0.7], [3, 0.2] ] )

def make51p():
    return makeGridSim(5, 1, [[],[],[[2,0]],[]], [0,0],
                       perfectSensorModel, perfectMotionModel) 

def make51ns():
    return makeGridSim(5, 1, [[], [], [[2,0]], []], [0,0],
                       noisySensorModel, perfectMotionModel)

def make51nm():
    return makeGridSim(5, 1, [[], [], [[2,0]], []], [0,0],
                       perfectSensorModel, noisyMotionModelX)

def make51nn():
    return makeGridSim(5, 1, [[], [], [[2,0]], []], [0,0],
                       noisySensorModel, noisyMotionModelX)

def make55():
    return makeGridSim(5, 5,
                       [[[2, 3], [0, 4]],[[1,1],[2,2]],[[2,0]], [[4,4]]], [0,0],
                       noisySensorModel, noisyMotionModel)

def make201pp():
    return makeGridSim(20, 1,
                       [[], [], [[8,0],[9,0],[10,0],[11,0],[12,0]], []], [0,0],
                       perfectSensorModel, perfectMotionModel) 
